package com.abcseo.ws2300;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPConnectionClosedException;
import org.apache.commons.net.ftp.FTPReply;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.exception.ParseErrorException;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.velocity.tools.generic.NumberTool;

/**
 * @author David George
 * @version $Id: $
 * 
 * This thread generates HTML pages from templates and publishes them to the
 * built-in HTTP server's htdocs directory and to an FTP server. A separate
 * image FTP server can also be specified.
 * 
 * Copyright 2002 David George (The Author), France, All rights reserved.
 */
public class PublishThread extends Thread implements WSThread {
	Configuration config;
	boolean running;
	WS2300 weatherStation;
	TemperatureHumidity temperature;
	Wind wind;
	Rain rain;
	Pressure pressure;
	private Logger logger = Logger.getLogger("com.abcseo");
	String server, username, password, directory;
	String iServer, iUsername, iPassword, iDirectory;
	String docRoot;

	/**
	 * 
	 * @param weatherStation
	 *            A weather station object
	 */
	PublishThread(WS2300 weatherStation, String docRoot) {
		this.weatherStation = weatherStation;
		temperature = weatherStation.getTemperature();
		wind = weatherStation.getWind();
		rain = weatherStation.getRain();
		pressure = weatherStation.getPressure();
		running = true;

		try {
			config = Configuration.getInstance();
		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"Cannot get configuration, exiting program " + e);
			System.exit(-1);
		}

		this.docRoot = docRoot;
		server = config.getFTPServer();
		username = config.getFTPUsername();
		password = config.getFTPPassword();
		directory = config.getFTPDirectory();

		if ((iServer = config.getFTPImageServer()) != null) {
			iPassword = config.getFTPImagePassword();
			iUsername = config.getFTPImageUsername();
			iDirectory = config.getFTPImageDirectory();
		}
	}

	/**
	 * Tempates are updated based on a configuration parameter. Files are pushed
	 * to the FTP server on some multiple of the template update period.
	 */
	public void run() {
		int timeout = config.getWebUpdate();
		int ftpTimeout = 0;

		logger.log(Level.INFO, "Starting Web and FTP Publishing");
		while (running) {
			processWeatherData();

			if ((ftpTimeout -= timeout) <= 0) {
				publishWeatherData();
				ftpTimeout = config.getFTPUpdate();
			}

			try {
				Thread.sleep(timeout);
			} catch (InterruptedException e) {
				logger.log(Level.INFO, "Publishing Thread Interupted");
			}
		}

	}

	/**
	 * Publish all .htm and .jpg files in the htdocs folder to the ftp folder
	 */
	private void publishWeatherData() {
		FilenameFilter filter;

		if (iServer != null) {
			/*
			 *  Image server is configured, upload JPEG files
			 */
			filter = new FilenameFilter() {
				public boolean accept(File dir, String name) {
					if (name.endsWith(".jpg") ) {
						return true;
					}

					return false;
				}
			};
			ftpConnection(iServer, iUsername, iPassword, iDirectory, filter);
			
			/*
			 * Upload .htm to ftp server
			 */
			filter = new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.endsWith(".htm");
				}
			};
			ftpConnection(server, username, password, directory, filter);

			
		} else if (server != null) {
			/*
			 * No image server but ftp server is configured, upload all HTML and JPEG files here
			 */
			filter = new FilenameFilter() {
				public boolean accept(File dir, String name) {
					if (name.endsWith(".htm") || name.endsWith(".jpg")) {
						return true;
					}

					return false;
				}
			};
			ftpConnection(server, username, password, directory, filter);
		}
	}

	/*
	 * Publish all .htm and .jpg files in the htdocs folder to the ftp folder
	 * 
	 * @param server @param username @param password @param directory @param
	 * filter
	 */
	private void ftpConnection(String server, String username, String password,
			String directory, FilenameFilter filter) {
		FTPClient ftp = new FTPClient();
		// add configuration parameter for timeout
		// ftp.setDefaultTimeout(10);

		try {
			int reply;

			ftp.connect(server);
			ftp.login(username, password);
			reply = ftp.getReplyCode();

			if (!FTPReply.isPositiveCompletion(reply)) {
				ftp.disconnect();
				logger.log(Level.WARNING, "FTP server " + server
						+ " refused connection");
				return;
			}

			ftp.changeWorkingDirectory(directory);
			ftp.setFileType(FTP.BINARY_FILE_TYPE);
			ftp.enterLocalPassiveMode();

			File dir = new File(docRoot);

			File[] files = dir.listFiles(filter);
			if (files == null) {
				logger.log(Level.WARNING, "no files to upload in "
						+ docRoot);
			} else {
				for (int i = 0; i < files.length; i++) {

					File f = (File) files[i];
					InputStream input = new FileInputStream(f);

					if (!ftp.storeFile(f.getName(), input)) {
						logger.log(Level.WARNING, "error uploading "
								+ f.getName() + " to " + server);
					}
					input.close();
				}
			}
			ftp.logout();
			ftp.disconnect();
		} catch (FileNotFoundException e) {
			logger.log(Level.WARNING, "file not found " + e);
		} catch (FTPConnectionClosedException e) {
			logger.log(Level.WARNING, "FTP Server: " + server
					+ " closed connection.");
		} catch (IOException e) {
			logger.log(Level.WARNING, e.toString());
		} finally {
			if (ftp.isConnected()) {
				try {
					ftp.disconnect();
				} catch (IOException e) {
					logger.log(Level.WARNING, e.toString());
				}
			}
		}
	}

	/**
	 * processWeatherData
	 */
	private void processWeatherData() {
		Configuration config = Configuration.getInstance();

		/*
		 * setup
		 */
		Properties p = new Properties();
		p.setProperty(Velocity.FILE_RESOURCE_LOADER_PATH, docRoot);
		p.setProperty("runtime.log", docRoot + "velocity.log");
		try {
			Velocity.init(p);
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Problem initializing velocity " + e);
			return;
		}
		/*
		 * Make a context object and populate with the data. This is where the
		 * Velocity engine gets the data to resolve the references (ex. $list)
		 * in the template
		 */
		VelocityContext context = new VelocityContext();
		NumberTool numberTool = new NumberTool();
		context.put("number", numberTool);
		context.put("config", config);
		context.put("temperature", temperature);
		context.put("wind", wind);
		context.put("rain", rain);
		context.put("pressure", pressure);

		/*
		 * get the Template object. This is the parsed version of your template
		 * input file. Note that getTemplate() can throw
		 * ResourceNotFoundException : if it doesn't find the template
		 * ParseErrorException : if there is something wrong with the VTL
		 * Exception : if something else goes wrong (this is generally
		 * indicative of as serious problem...)
		 */
		File dir = new File(config.getDocRoot());

		// exclude velocity template files from upload
		FilenameFilter filter = new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.endsWith(".vm");
			}
		};

		String[] files = dir.list(filter);
		if (files == null) {
			logger.log(Level.INFO, "no files to upload in " + docRoot);
		} else {
			for (int i = 0; i < files.length; i++) {
				Template template = null;

				try {
					template = Velocity.getTemplate(files[i]);
				} catch (ResourceNotFoundException rnfe) {
					logger.log(Level.WARNING, " cannot find template "
									+ files[i]);
					continue;
				} catch (ParseErrorException pee) {
					logger.log(Level.WARNING, "Syntax error in template "
							+ files[i] + ":" + pee);
					continue;
				} catch (Exception e) {
					logger.log(Level.WARNING, "General Velocity Exception " + e);
					continue;
				}

				/*
				 * Now have the template engine process your template using the
				 * data placed into the context. Think of it as a 'merge' of the
				 * template and the data to produce the output stream.
				 */
				String outfile = files[i].substring(0, files[i]
						.lastIndexOf('.'));
				outfile = docRoot + outfile + ".htm";

				logger.log(Level.FINE, "writing to " + outfile);
				OutputStream htmlStream = null;
				try {
					htmlStream = new FileOutputStream(outfile);
				} catch (FileNotFoundException e) {
					logger.log(Level.WARNING, "File Not Found " + outfile + " "
							+ e);
					continue;
				}

				BufferedWriter writer = writer = new BufferedWriter(
						new OutputStreamWriter(htmlStream));

				try {
					if (template != null) {
						template.merge(context, writer);
					}
				} catch (Exception e) {
					logger.log(Level.WARNING, "Problem merging " + outfile
							+ " " + e);
					continue;
				}

				try {
					writer.flush();
					writer.close();
				} catch (Exception e) {
					logger.log(Level.WARNING, "Problem closing " + outfile
							+ " " + e);

				}
			}// for
		}
	}

	/**
	 * Cleanup
	 */
	public void shutdown() {
		this.interrupt();
		running = false;

		logger.log(Level.INFO, "Publisher Shutdown");
	}
}
