/*
 * Created on 12-May-2005
 *
 */
package com.abcseo.ws2300;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.time.Minute;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.xy.IntervalXYDataset;

/**
 * Set of methods related to temperature * Temperature is 4 digits e.g.: 17.3 C
 * held in two bytes BCD format + an offset of 30. e.g. 47 30
 * 
 * @author david
 * 
 * Changes
 * 3 May 2006: Petr Brouzda
 * Fix to SQL Select for Rain values
 */
public class Rain extends Thread implements WSThread {
	WS2300 weatherStation;
	public static int RAIN_TOTAL = 0x4D2;
	double totalRainfall;

	boolean SI;
	long time;
	private Logger logger = Logger.getLogger("com.abcseo");
	Configuration config;
	boolean running;
	int refreshTime;
	Connection c;
	PreparedStatement rainIns, rainSel;

	private Rain() {
	};

	/**
	 * Default constructor, SI (metric) units are used
	 * 
	 * @param weatherStation
	 */
	public Rain(WS2300 weatherStation) {
		this.weatherStation = weatherStation;
		config = Configuration.getInstance();
		refreshTime = config.getRainRefresh();
		try {
			c = DriverManager.getConnection(config.getJDBCDatabase(), config
					.getJDBCUsername(), config.getJDBCPassword());
			rainIns = c.prepareStatement("INSERT INTO Rain VALUES (?,?)");
			rainSel = c.prepareStatement("SELECT * FROM Rain WHERE TIME > ? ORDER BY TIME ASC");
			// initialise object with most recent values from database
			Statement s = c.createStatement();
			s.execute("SELECT TOP 1 * FROM Rain ORDER BY TIME DESC");
			ResultSet rs = s.getResultSet();

			if (rs.next()) {
				totalRainfall = rs.getDouble(1);
				time = rs.getTimestamp(2).getTime();
			} else {
				totalRainfall = 0.0;
				time = 0; // the epoch
			}
		} catch (SQLException e) {
			logger.log(Level.SEVERE,
					"Critical Database error while initialising Rain Object "
							+ e);
			System.exit(-1);
		}
	}

	/*
	 * 
	 * Returns: rain (double) converted to unit given in config
	 */
	public double getTotal() {
		return totalRainfall;
	}

	/**
	 * Update the current wind speed and direction values from the weather
	 * station
	 * 
	 * @author David George
	 * @author Open2300
	 */
	private void update(long time) throws WSException, SQLException {
		int data[];
		data = weatherStation.read(RAIN_TOTAL, 3);

		double r = (((data[2] >> 4) * 1000 + (data[2] & 0xF) * 100
				+ (data[1] >> 4) * 10 + (data[1] & 0xF) + (data[0] >> 4) / 10.0 + (data[0] & 0xF) / 100.0));

		/*
		 * record only changes, if the figure changes then we need to look at
		 * the last figure, if this was the further back then the update time
		 * insert a record t - update with the previous figure into the database
		 * first... same for wind and temperature - this probably wants to go
		 * into update method
		 */
		if (r != totalRainfall) {
			totalRainfall = r;
			update(totalRainfall, time);
		}
		
		logger.log(Level.FINE, "Total fainfall is: %.2f mm\n", totalRainfall);
		return;
	}

	/**
	 * Insert a windspeed record into the database. This does not set the
	 * current time.
	 * 
	 * @return if the record is older than the current record return false
	 */
	boolean update(double rainfall, long recordTime) throws SQLException {

		if (recordTime < time) {
			return false;
		}

		rainIns.setDouble(1, rainfall);
		rainIns.setTimestamp(2, new java.sql.Timestamp(recordTime));
		rainIns.execute();

		return true;
	}

	public boolean resetTotal() {
		return true;

	}

	/**
	 * Main collector thread, talks to the weather station and writes data to an
	 * RDBMS
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		running = true;
		logger.log(Level.INFO, "Starting Rainfall Collection Thread");

		while (running) {
			try {
				update(weatherStation.getCurrentTimeMillis());
				createChart();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "error writing rainfall to database "
						+ e);
			} catch (WSException wse) {
				logger.log(Level.SEVERE,
						"error reading rainfall from weather station " + wse);
			}

			try {
				sleep(refreshTime);
			} catch (InterruptedException e) {
			}
		} // while forever
		logger.log(Level.INFO, "Stopping Rainfall Collection Thread");
	}

	public void shutdown() {
		running = false;
		this.interrupt();

		try {
			rainIns.close();
			rainSel.close();
			Statement stmt = c.createStatement();
			c.close();
		} catch (SQLException e) {
			logger.log(Level.WARNING, "error closing connection to Rain table "
					+ e);
		}
	}

	private void createChart() {
		java.sql.Timestamp tstamp = new java.sql.Timestamp(System
				.currentTimeMillis() - 86000000);

		boolean firstTime = true;
		double base = 0.0;

		try {
			rainSel.setTimestamp(1, tstamp);
			ResultSet rs = rainSel.executeQuery();

			TimeSeries series = new TimeSeries("Total Rainfall",
					org.jfree.data.time.Minute.class);

			while (rs.next()) {
				double rain = round(rs.getDouble(1), 1);
				tstamp = rs.getTimestamp(2);
				if (firstTime) {
					base = rain;
					firstTime = false;
					continue;
				}
				rain = rain - base;
				Minute m = new Minute(tstamp);
				series.addOrUpdate(m, rain);
			}

			IntervalXYDataset data = new TimeSeriesCollection(series);
			JFreeChart chart = ChartFactory.createXYBarChart(
					"Rainfall Total (24-Hour)", "Time", true, "Rainfall (mm)",
					data, PlotOrientation.VERTICAL, true, false, false);
			XYPlot xyplot = (XYPlot) chart.getPlot();
			DateAxis dateaxis = (DateAxis) xyplot.getDomainAxis();
			dateaxis.setDateFormatOverride(new SimpleDateFormat("kk:mm"));

			try {
				ChartUtilities.saveChartAsJPEG(new java.io.File(config
						.getDocRoot()
						+ "Total-Rainfall-24hr.jpg"), chart, 500, 300);
			} catch (java.io.IOException exc) {
				logger.log(Level.WARNING, "Error writing image to file "
						+ config.getDocRoot() + "Total-Rainfall-24hr.jpg");
			}

		} catch (SQLException e) {
			logger.log(Level.SEVERE, "error reading rain data from database " + e);
		}
	}

	/**
	 * Round a double value to a specified number of decimal places.
	 * 
	 * @param val
	 *            the value to be rounded.
	 * @param places
	 *            the number of decimal places to round to.
	 * @return val rounded to places decimal places.
	 */
	public double round(double val, int places) {
		long factor = (long) Math.pow(10, places);

		// Shift the decimal the correct number of places
		// to the right.
		val = val * factor;

		// Round to the nearest integer.
		long tmp = Math.round(val);

		// Shift the decimal the correct number of places
		// back to the left.
		return (double) tmp / factor;
	}
}