/*
 * Created on 12-May-2005
 *
 */
package com.abcseo.ws2300;

import java.awt.Color;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CompassFormat;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.axis.TickUnits;
import org.jfree.chart.plot.PolarPlot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.DefaultPolarItemRenderer;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.time.Minute;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * Set of methods related to temperature * Temperature is 4 digits e.g.: 17.3 C
 * held in two bytes BCD format + an offset of 30. e.g. 47 30
 * 
 * @author david
 * 
 */
public class Wind extends Thread implements WSThread {
	// addresses in the WS2300
	private static int WIND = 0x527;
	private static int MAXWINDRETRIES = 3;
	private static int AVERAGE_TRIES = 1;

	WS2300 weatherStation;
	boolean SI;
	double currentDirection;
	double currentSpeed;
	double averageSpeed;
	double averageDirection;
	double gust;
	long time;
	private Logger logger = Logger.getLogger("com.abcseo");
	Configuration config;
	boolean running;
	int refreshTime;
	Connection c;
	PreparedStatement windIns, windSel;
	// used for averaging last five wind speed measurements
	private int counter = 1;

	private Wind() {
	};

	/**
	 * Default constructor, SI (metric) units are used
	 * 
	 * Get's the last windspeed/direction reading from the database
	 * 
	 * @param weatherStation
	 */
	Wind(WS2300 weatherStation) {
		this.weatherStation = weatherStation;

		config = Configuration.getInstance();
		refreshTime = config.getWindRefresh();

		// create connection to the database
		try {
			Class.forName(config.getJDBCDriver());
		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"ERROR: failed to load JDBC driver in Wind object");
			System.exit(-1);
		}

		try {
			c = DriverManager.getConnection(config.getJDBCDatabase(), config
					.getJDBCUsername(), config.getJDBCPassword());
			windIns = c.prepareStatement("INSERT INTO Wind VALUES (?,?,?,?)");
			windSel = c.prepareStatement("SELECT * FROM Wind WHERE TIME > ?");

			// initialise object with most recent values from database
			Statement s = c.createStatement();
			s.execute("SELECT TOP 1 * FROM Wind ORDER BY TIME DESC");
			ResultSet rs = s.getResultSet();

			if (rs.next()) {
				averageSpeed = rs.getDouble(1);
				averageDirection = rs.getDouble(2);
				gust = rs.getDouble(3);
				time = rs.getTimestamp(4).getTime();
			} else {
				averageSpeed = 0.0;
				averageDirection = 0;
				gust = 0.0;
				time = 0; // the epoch
			}
		} catch (SQLException e) {
			logger.log(Level.SEVERE,
					"Critical Database error while initialising Wind Object "
							+ e);
			System.exit(-1);
		}
	}

	/**
	 * Return the current average windspeed in km/h
	 */
	public double getSpeed() {
		return averageSpeed * 3.6;
	}

	/**
	 * Return the current wind direction
	 */
	public double getDirection() {
		return averageDirection;
	}

	/**
	 * Convert a wind direction reading into a compass bearing
	 * 
	 * @returns Wind direction
	 */
	public String getCompassDirection() {
		String compass = "N";

		if (averageDirection > 11.25) {
			compass = "NNE";
			if (averageDirection > 33.75) {
				compass = "NE";
				if (averageDirection > 56.25) {
					compass = "ENE";
					if (averageDirection > 78.75) {
						compass = "E";
						if (averageDirection > 101.25) {
							compass = "ESE";
							if (averageDirection > 123.75) {
								compass = "SE";
								if (averageDirection > 146.25) {
									compass = "SSE";
									if (averageDirection > 168.75) {
										compass = "S";
										if (averageDirection > 191.25) {
											compass = "SSW";
											if (averageDirection > 213.75) {
												compass = "SW";
												if (averageDirection > 236.25) {
													compass = "WSW";
													if (averageDirection > 258.75) {
														compass = "W";
														if (averageDirection > 281.25) {
															compass = "WNW";
															if (averageDirection > 303.75) {
																compass = "NW";
																if (averageDirection > 326.25) {
																	compass = "NNW";
																	if (averageDirection > 348.75) {
																		compass = "N";
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		return compass;
	}

	/**
	 * Update the current wind speed and direction values from the weather
	 * station
	 * 
	 * @author David George
	 * @author Open2300
	 */
	private void update(long time) throws WSException, SQLException {
		int data[];

		for (int i = 0; i < MAXWINDRETRIES; i++) {
			data = weatherStation.read(WIND, 3);

			// Invalid wind data
			if ((data[0] != 0x00)
					|| ((data[1] == 0xFF) && (((data[2] & 0xF) == 0) || ((data[2] & 0xF) == 1)))) {
				// back off 10 seconds
				try {
					sleep(10000);
				} catch (InterruptedException e) {
				}
				continue;
			} else {
				// Calculate wind directions

				currentDirection += (data[2] >> 4) * 22.5;
				// Calculate raw wind speed
				double speed = (((data[2] & 0xF) << 8) + (data[1])) / 10.0;
				if (speed > gust) {
					gust = speed;
				}
				currentSpeed += speed;

				if (counter == AVERAGE_TRIES) {
					averageSpeed = currentSpeed / AVERAGE_TRIES;
					averageDirection = currentDirection / AVERAGE_TRIES;
					update(averageSpeed, averageDirection, gust, time);
					currentSpeed = 0.0;
					currentDirection = 0.0;
					counter = 1;
				} else {
					counter++;
				}
				return;
			}
		}// for

		logger.log(Level.WARNING,
				"could not get reliable windspeed reading after  "
						+ MAXWINDRETRIES + " retries");
		return;
	}

	/**
	 * Insert a windspeed record into the database. This does not set the
	 * current time.
	 * 
	 * @return false if the record is older than the current record in the
	 *         database
	 */
	boolean update(double speed, double direction, double gust, long recordTime)
			throws SQLException {

		if (recordTime < time) {
			return false;
		}

		/*
		 * Don't add zero records to the database ???
		 */
		if (gust > 0.0) {
			windIns.setDouble(1, speed);
			windIns.setDouble(2, direction);
			windIns.setDouble(3, gust);
			windIns.setTimestamp(4, new java.sql.Timestamp(recordTime));
			windIns.execute();
		}

		return true;
	}

	/**
	 * Main collector thread, talks to the weather station and writes data to an
	 * RDBMS
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		running = true;
		logger.log(Level.INFO, "Starting Windspeed Collection Thread");

		while (running) {
			try {
				update(weatherStation.getCurrentTimeMillis());
				createChart();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "error writing to database " + e);
			} catch (WSException wse) {
				logger.log(Level.SEVERE, "error reading from weather station "
						+ wse);
			}

			try {
				sleep(refreshTime);
			} catch (InterruptedException e) {
			}
		} // while forever
		logger.log(Level.INFO, "Stopping Windspeed Collection Thread");
	}

	public void shutdown() {
		running = false;
		this.interrupt();

		try {
			windIns.close();
			windSel.close();
			Statement stmt = c.createStatement();
			c.close();
		} catch (SQLException e) {
			logger.log(Level.WARNING, "error closing connection to Wind table "
					+ e);
		}
	}

	/*
	 * * wind_all Read wind speed, wind direction and last 5 wind directions
	 * 
	 * Input: Handle to weatherstation wind_speed_conv_factor controlling
	 * convertion to other units than m/s
	 * 
	 * Output: winddir_index Current wind direction expressed as ticks from
	 * North where North=0. Used to convert to direction string winddir Array of
	 * doubles containing current winddirection in winddir[0] and the last 5 in
	 * the following positions all given in degrees
	 * 
	 * Returns: Wind speed (double) in the unit given in the loaded config
	 */
	public void getAll() {

	}

	/*
	 * * wind_minmax Read min/max wind speeds with timestamps
	 * 
	 * Input: Handle to weatherstation wind_speed_conv_factor controlling
	 * convertion to other units than m/s
	 * 
	 * Output: Wind wind_min and wind_max (double) unit defined by config
	 * conversion factor Timestamps for wind_min and wind_max in pointers to
	 * timestamp structures for time_min and time_max
	 * 
	 * Returns: wind max (double)
	 * 
	 * Note: The function is made so that if a pointer to wind_min/max and
	 * time_min/max is a NULL pointer the function ignores this parameter.
	 * Example: if you only need wind_max use the function like this.. windmax =
	 * wind_minmax(ws2300,METERS_PER_SECOND,NULL,NULL,NULL,NULL);
	 */
	public void getMinMax() {
	}

	/*
	 * wind_reset Reset min/max wind with timestamps
	 * 
	 * Input: Handle to weatherstation minmax - char (8 bit integer) that
	 * controls if minimum, maximum or both are reset Output: None
	 * 
	 * Returns: 1 if success @return
	 */
	public boolean reset() {
		return false;
	}

	/*
	 * * windchill Read wind chill, current value only
	 * 
	 * Input: Handle to weatherstation temperature_conv flag (integer)
	 * controlling convertion to deg F
	 * 
	 * Returns: wind chill (deg C if config.temperature_conv is not set) (deg F
	 * if config.temperature_conv is set)
	 * 
	 * It is recommended to run this right after a wind speed reading to enhance
	 * the likelyhood that the wind speed is valid
	 */
	public double getWindChill() {
		return 0.0;
	}

	/*
	 * * windchill_minmax Read wind chill min/max with timestamps
	 * 
	 * Input: Handle to weatherstation temperature_conv flag (integer)
	 * controlling convertion to deg F
	 * 
	 * Output: Windchill wc_min and wc_max (deg C if config.temperature_conv is
	 * not set) (deg F if config.temperature_conv is set) Timestamps for wc_min
	 * and wc_max in pointers to timestamp structures for time_min and time_max
	 * 
	 * Returns: Nothing
	 */
	public double getWindChillMinMax() {
		return 0.0;
	}

	/*
	 * * windchill_minmax Read wind chill min/max with timestamps
	 * 
	 * Input: Handle to weatherstation temperature_conv flag (integer)
	 * controlling convertion to deg F
	 * 
	 * Output: Windchill wc_min and wc_max (deg C if config.temperature_conv is
	 * not set) (deg F if config.temperature_conv is set) Timestamps for wc_min
	 * and wc_max in pointers to timestamp structures for time_min and time_max
	 * 
	 * Returns: Nothing
	 * 
	 */
	public boolean windChillReset() {
		return true;
	}

	private void createChart() {
		java.sql.Timestamp tstamp = new java.sql.Timestamp(System
				.currentTimeMillis() - 86000000);

		try {
			windSel.setTimestamp(1, tstamp);
			ResultSet rs = windSel.executeQuery();

			XYSeries series = new XYSeries("Wind Speed (km/h)");
			TimeSeries ws_series = new TimeSeries("Wind Speed (km/h)",
					org.jfree.data.time.Minute.class);
			TimeSeries wd_series = new TimeSeries("Wind Direction",
					org.jfree.data.time.Minute.class);

			while (rs.next()) {
				double speed = rs.getDouble(1) * 3.6;
				double direction = rs.getDouble(2);
				tstamp = rs.getTimestamp(4);

				series.addOrUpdate(direction, speed);

				Minute m = new Minute(tstamp);
				ws_series.addOrUpdate(m, speed);
				wd_series.addOrUpdate(m, direction);
			}

			// polar chart
			XYDataset data = new XYSeriesCollection(series);
			JFreeChart chart = ChartFactory.createPolarChart(
					"Wind Trend (24 hour)", data, true, false, false);
			
			PolarPlot plot = (PolarPlot) chart.getPlot();
			DefaultPolarItemRenderer r = new DefaultPolarItemRenderer();
			
			r.setSeriesFilled(0, true);
			r.setSeriesPaint(0, Color.blue);
            plot.setRenderer(r);

			try {
				ChartUtilities.saveChartAsJPEG(new java.io.File(config
						.getDocRoot()
						+ "Prevailing-24hr.jpg"), chart, 500, 300);
			} catch (java.io.IOException exc) {
				logger.log(Level.WARNING, "Error writing image to file "
						+ config.getDocRoot() + "Prevailing-24hr.jpg");
			}

			// Compass chart
			TimeSeriesCollection windData = new TimeSeriesCollection();
			windData.addSeries(wd_series);

			chart = ChartFactory.createTimeSeriesChart(
					"Wind Speed and Direction", "Time", "Direction", windData,
					true, false, false);

			XYPlot xyplot = (XYPlot) chart.getPlot();
			xyplot.getDomainAxis().setLowerMargin(0.0D);
			xyplot.getDomainAxis().setUpperMargin(0.0D);
			NumberAxis numberaxis = (NumberAxis) xyplot.getRangeAxis();
			numberaxis.setAutoRangeIncludesZero(false);
			TickUnits tickunits = new TickUnits();
			tickunits.add(new NumberTickUnit(180D, new CompassFormat()));
			tickunits.add(new NumberTickUnit(90D, new CompassFormat()));
			tickunits.add(new NumberTickUnit(45D, new CompassFormat()));
			tickunits.add(new NumberTickUnit(22.5D, new CompassFormat()));
			numberaxis.setStandardTickUnits(tickunits);
			xyplot.setRangeAxis(numberaxis);

			XYAreaRenderer xyarearenderer = new XYAreaRenderer();
			NumberAxis numberaxis1 = new NumberAxis("Speed");
			numberaxis1.setRange(0.0D, 180D);
			xyplot.setRangeAxis(1, numberaxis);
			xyarearenderer.setSeriesPaint(0, new Color(0, 0, 255, 128));

			TimeSeriesCollection speedData = new TimeSeriesCollection();
			speedData.addSeries(ws_series);
			xyplot.setDataset(1, speedData);
			xyplot.setRenderer(1, xyarearenderer);
			xyplot.setRangeAxis(1, numberaxis1);
			xyplot.mapDatasetToRangeAxis(1, 1);

			try {
				ChartUtilities.saveChartAsJPEG(new java.io.File(config
						.getDocRoot()
						+ "Wind-24hr.jpg"), chart, 500, 300);
			} catch (java.io.IOException exc) {
				logger.log(Level.WARNING, "Error writing image to file "
						+ config.getDocRoot() + "Wind-24hr.jpg");
			}
		} catch (SQLException e) {
			logger.log(Level.SEVERE, "error reading wind data from database "
					+ e);
		}
	}
}